module Combine (
  Mean(..)
, Rank
, Weight
, combine
, main
) where

import           Control.Arrow (second)
import           Control.Parallel.Strategies (parMap, rdeepseq)
import qualified Data.ByteString.Char8 as BC
import           Data.Foldable (toList)
import           Data.IntMap (IntMap)
import qualified Data.IntMap as IM
import           Data.List (intercalate, sort, transpose)
import qualified Data.Set as Set
import           Data.Tuple (swap)
import           Options.Applicative
import           Options.Applicative.Help.Chunk
import           System.Environment (getArgs, getProgName)
import           System.Exit (exitFailure)
import           System.IO (hPutStrLn, stderr)

import           Interpreter
import           Types
import           TypeUtils


-- -----------------------------------------------------------------------------
-- Combination of predictions

type PredictionID = Int
type Rank         = Int
type Weight       = Double

data Mean = Arithmetic | Geometric | Quadratic | Harmonic | Maximum
  deriving (Read, Show)

meanFunction :: (Floating a, Ord a) => Mean -> [(a, a)] -> a
meanFunction Arithmetic =         sum . map (uncurry (*))
meanFunction  Geometric =         product . map (uncurry (**))
meanFunction  Quadratic =         sum . map (\ (x, w) -> x * x * w)
meanFunction   Harmonic = (1 /) . sum . map (\ (x, w) -> w / x)
meanFunction    Maximum =   maximum . map (uncurry (*))

labelRankMap :: [Labels] -> IntMap [(PredictionID, Rank)]
labelRankMap lblss =
  foldl (\ m (i, lbls) -> foldr (add i) m lbls) IM.empty zipped
  where zipped = zip [1 ..] $ map (zip [1 ..]) lblss
        add i (r, l) = IM.insertWith (++) l [(i, r)]

fillRanks :: PredictionID -> Rank -> [(PredictionID, Rank)] -> [Rank]
fillRanks maxI maxRank xs = map snd $ sort $ xs ++ remaining
  where remaining = [(i, maxRank) | i <- [1 .. maxI], i `notElem` map fst xs]

combine :: Mean -> Rank -> [Weight] -> [Labels] -> Labels
combine mean maxRank weights lblss = map snd $ sort $ map swap meaned
  where m = IM.toList $ labelRankMap lblss
        filled   = map (second $ fillRanks (length weights) maxRank) m
        weighted = map (second $ flip zip weights . map fromIntegral) filled
        meaned   = map (second $ meanFunction mean) weighted


-- -----------------------------------------------------------------------------
-- Main function

usageInfo :: String -> String
usageInfo progName = intercalate "\n"
  [ "Usage: " ++ progName ++ " <seq> <eval> <mean> <maxRank> "
    ++ "[<file1> <weight1> [... [<file_n> <weight_n>]]]"
  , "Combine and output predictions using a voting scheme."
  , ""
  , "The predicted usefulness of a given premise is computed using a "
  , "weighted average of the usefulnesses from the combined files."
  , ""
  , "The possible averages are:"
  , "  Arithmetic, Geometric, Quadratic, Maximum, or Harmonic (typically best)"
  , ""
  , "A prediction file specifies maximally n predictions for each theorem."
  , "The votes for the first n are their number in the sequence, and"
  , "for the rest, we consider their vote to be the given number `maxRank'."
  , ""
  , "Example usage:"
  , "  " ++ progName ++ " seq eval Harmonic 1024 file1 0.3 file2 0.4 file3 0.3"
  ]

data Arguments = Arguments {
  sequenceFile :: FilePath
, evaluationFile :: FilePath
, meanMethod :: Mean
, maximalRank :: Int
, fileWeights :: [(FilePath, Double)]
} deriving Show


helpChunk = helpDoc . Just . extractChunk

helpUnlines = helpChunk . vcatChunks . map paragraph

averageDoc =
  [ "average"
  , "Options: Arithmetic, Geometric, Quadratic, Maximum, or Harmonic (typically best)"
  ]

fileWeightParser = (,)
  <$> argument str  (metavar "file")
  <*> argument auto (metavar "weight")

argumentParser = Arguments
  <$> argument str  (metavar "seq"  <> help "sequence file")
  <*> argument str  (metavar "eval" <> help "evaluation file")
  <*> argument auto (metavar "mean" <> helpUnlines averageDoc)
  <*> argument auto (metavar "maxRank")
  <*> many fileWeightParser

parserInfo = info (helper <*> argumentParser)
  ( fullDesc
  <> progDesc "Combine and output predictions using a voting scheme."
  )

showUsage :: IO ()
showUsage = do
  progName <- getProgName
  hPutStrLn stderr $ usageInfo progName

-- | Interpret a sequence of path-weight pairs, such as "file1 0.3 file2 0.4".
parsePreds :: [String] -> [(FilePath, Weight)]
parsePreds (path : weight : xs) = (path, read weight) : parsePreds xs
parsePreds [_] = error "prediction weight not specified"
parsePreds [ ] = []


main :: IO ()
main = do
  args <- getArgs
  (seqFP, evalFP, mean, maxRank, predictions) <- case args of
    (sF : eF : m : mR : p) -> return (sF, eF, read m, read mR, parsePreds p)
    _ -> showUsage >> exitFailure
{-
  Arguments seqFP evalFP mean maxRank predictions <- execParser parserInfo
-}

  let (predFPs, predWeights) = unzip predictions
      combineF = take maxRank . combine mean maxRank predWeights

  lblMap        <- loadSequenceFile seqFP
  eval          <- Set.fromList `fmap` loadEvaluationFile lblMap evalFP
  fileLineLabel <- mapM (loadLabelsFile lblMap) predFPs

  let lineFileLabel = zip (map SampleID [0 ..]) $ labelTranspose fileLineLabel
      evaluated = [(i, l) | (i, l) <- lineFileLabel, i `Set.member` eval]
      combined = parMap rdeepseq (second combineF) evaluated
  mapM_ (BC.putStrLn . uncurry (showSampleLabels lblMap)) combined

  where labelTranspose = transpose . map (\ (SampleLabels lT) -> toList lT)
