module Interpreter
( loadSequenceFile
, loadLabels
, loadLabelsFile
, loadFeaturesFile
, loadFeaturesLine
, loadEvaluationFile
, loadDotFile
, loadFastXMLPredictions
, loadFastXMLPredictionsFile
) where

import           Control.Monad (liftM)
import           Data.ByteString (ByteString)
import qualified Data.ByteString as B
import qualified Data.ByteString.Char8 as BC
import           Data.List (sortBy)
import           Data.Maybe (mapMaybe, fromMaybe)
import           Data.Ord (comparing)
import           Data.Sequence (Seq)
import qualified Data.Sequence as Seq

import qualified IntBiMap as IBM
import           Parser
import           Types


type BBList = (ByteString, [ByteString])


-- -----------------------------------------------------------------------------
-- Error messages

unknownLabelMsg :: ByteString -> String
unknownLabelMsg lbl = "unknown label " ++ BC.unpack lbl ++ " encountered"

unknownFeatureMsg :: ByteString -> String
unknownFeatureMsg ftr = "unknown feature " ++ BC.unpack ftr ++ " encountered"


-- -----------------------------------------------------------------------------
-- Interpretation

interpretSequence :: [ByteString] -> LabelBM
interpretSequence = foldl f IBM.empty
  where
    f lblMap lbl = case IBM.lookupInt lblMap lbl of
      Nothing -> IBM.insert lbl lblMap
      Just _ -> error $ "label " ++ BC.unpack lbl ++ " encountered twice"

interpretLabels :: LabelBM -> [BBList] -> Seq Labels
interpretLabels lblMap = foldl f (Seq.replicate (IBM.size lblMap) [])
  where
    f lblss (lbl, lbls) =
      let lblsi = foldr handleLabel [] lbls in
      case IBM.lookupInt lblMap lbl of
        Just i -> Seq.update i lblsi lblss
        Nothing -> error $ unknownLabelMsg lbl

    handleLabel lbl lblsi =
      case IBM.lookupInt lblMap lbl of
        Just i -> i : lblsi
        Nothing -> error $ unknownLabelMsg lbl

loadLabels :: LabelBM -> [ByteString] -> SampleLabels
loadLabels lblMap = SampleLabels . interpretLabels lblMap . map parseLabelsLine


interpretFeaturesLine :: FeatureBM -> [ByteString] -> Features
interpretFeaturesLine ftrMap =
  map (\ f -> fromMaybe (error $ unknownFeatureMsg f) $ IBM.lookupInt ftrMap f)

interpretFeatures :: LabelBM -> [BBList] -> (Seq Features, FeatureBM)
interpretFeatures lblMap =
  foldl f (Seq.replicate (IBM.size lblMap) [], IBM.empty)
  where
    f (ftrss, ftrMap) (lbl, ftrs) =
      let (ftrsi, ftrMap') = foldr handleFeature ([], ftrMap) ftrs in
      case IBM.lookupInt lblMap lbl of
        Just i -> (Seq.update i ftrsi ftrss, ftrMap')
        Nothing -> error $ unknownLabelMsg lbl

    handleFeature ftr (ftrs, ftrMap) =
      case IBM.lookupInt ftrMap ftr of
        Nothing -> (IBM.nextInt ftrMap : ftrs, IBM.insert ftr ftrMap)
        Just i -> (i : ftrs, ftrMap)

loadFeatures :: LabelBM -> [ByteString] -> (Seq Features, FeatureBM)
loadFeatures lblMap = interpretFeatures lblMap . map parseFeaturesLine


-- -----------------------------------------------------------------------------
-- FastXML


loadFastXMLPredictionLine :: ByteString -> Labels
loadFastXMLPredictionLine =
  map fst . sortBy (flip $ comparing snd) . parseFastXMLPredictionsLine

mergeSamplesLabels :: Samples -> [Labels] -> SampleLabels
mergeSamplesLabels smpls = SampleLabels .
  foldr (\ (SampleID s, l) -> Seq.update s l)
    (Seq.replicate (maximum (map (\ (SampleID s) -> s) smpls) + 1) []) .
  zip smpls

loadFastXMLPredictions :: Samples -> [ByteString] -> SampleLabels
loadFastXMLPredictions eval =
  mergeSamplesLabels eval . map loadFastXMLPredictionLine


-- -----------------------------------------------------------------------------
-- I/O

readFileLines :: FilePath -> IO [ByteString]
readFileLines = liftM BC.lines . B.readFile

loadSequenceFile :: FilePath -> IO LabelBM
loadSequenceFile = liftM interpretSequence . readFileLines

loadLabelsFile :: LabelBM -> FilePath -> IO SampleLabels
loadLabelsFile lblMap = liftM (loadLabels lblMap) . readFileLines

loadFeaturesFile :: LabelBM -> FilePath -> IO (Seq Features, FeatureBM)
loadFeaturesFile lblMap = liftM (loadFeatures lblMap) . readFileLines

loadEvaluationFile :: LabelBM -> FilePath -> IO Samples
loadEvaluationFile lblMap = liftM (map (SampleID . IBM.findInt lblMap)) . readFileLines

loadFeaturesLine :: FeatureBM -> ByteString -> Features
loadFeaturesLine ftrMap = interpretFeaturesLine ftrMap . parseStringList

loadDotFile :: FilePath -> IO [(ByteString, ByteString)]
loadDotFile = liftM (mapMaybe parseDotLine) . readFileLines

loadFastXMLPredictionsFile :: Samples -> FilePath -> IO SampleLabels
loadFastXMLPredictionsFile s = liftM (loadFastXMLPredictions s) . readFileLines
