{-# LANGUAGE NoMonomorphismRestriction #-}

module Predict (main) where

import           Control.Arrow ((&&&))
import           Control.Monad (foldM, when)
import           Control.Monad.IO.Class (liftIO)
import qualified Data.Foldable as F
import           Data.List (intercalate)
import           Data.Maybe (isJust)
import qualified Data.ByteString.Char8 as BC
import           System.Console.GetOpt
import           System.Environment (getArgs, getProgName)
import           System.Exit (exitFailure)
import           System.IO (hPutStr, stderr)

import           Forest (initialiseForest)
import           KNN (initialiseKNN)
import           NaiveBayes (initialiseNaiveBayes)

import qualified Interpreter as I
import           Parser (parsePredictorOptions)
import           Predictor
import           Types
import           TypeUtils
import           Utils


interaction :: Predictor p => p -> Options -> PredictorT p IO ()
interaction predictor opts = do
  (fM, nAL) <- asksLData (featureBM &&& nAllLabels)

  p <- maybe (promotePredictor $ learn predictor $ samplesRange (SampleID 0) (SampleID nAL))
            (loadState predictor) $ optLoadState opts

  F.mapM_ (saveState p) (optSaveState opts)

  linesInteractM (promotePredictor . query p . I.loadFeaturesLine fM)

  where query p ftrs = predict p ftrs >>= showLabelsM . take nPreds
        nPreds = optNPredictions opts

evaluation :: Predictor p => FilePath -> p -> Options -> PredictorT p IO ()
evaluation evalF predictor opts = do
  lM <- asksLData labelBM
  smpls <- liftIO $ I.loadEvaluationFile lM evalF

  p <- maybe (return predictor) (loadState predictor) $ optLoadState opts

  p' <- foldM step p $ zip (SampleID 0 : smpls) smpls
  F.mapM_ (saveState p') (optSaveState opts)

  where
    learned = isJust $ optLoadState opts

    step :: Predictor p => p -> (SampleID, SampleID) -> PredictorT p IO p
    step p (from, to) = do
      fT <- asksLData sampleFeatures
      p' <- if learned then return p
                      else promotePredictor $ learn p $ samplesRange from to
      pr <- promotePredictor $ predict p' (getSampleFeatures fT to)
      let predictions = take (optNPredictions opts) pr
      lp <- promotePredictor $ showSampleLabelsM to predictions
      liftIO (BC.putStrLn lp)
      return p'


showStats :: LearnData -> String
showStats lD = unlines
  [ "Total number of labels: "     ++ show (nAllLabels lD)
  , "Total number of features: "   ++ show (nAllFeatures lD)
  , "Average number of labels: "   ++ show (lTAvg $ sampleLabels   lD)
  , "Average number of features: " ++ show (fTAvg $ sampleFeatures lD)
  ]
  where
    avg :: [[a]] -> Double
    avg = average . map (fromIntegral . length)
    tableAvg = avg . F.toList
    fTAvg (SampleFeatures fT) = tableAvg fT
    lTAvg (  SampleLabels lT) = tableAvg lT

featureHistogram :: SampleFeatures -> Histogram
featureHistogram (SampleFeatures fT) = histogram $ concat $ F.toList fT

writeFeatureHistogram :: FilePath -> SampleFeatures -> IO ()
writeFeatureHistogram fp = writeFile fp . show . featureHistogram


data Options = Options {
  optEvalFile :: Maybe FilePath
, optNPredictions :: Int
, optPredictor :: String
, optSaveState :: Maybe FilePath
, optLoadState :: Maybe FilePath
, optHistogram :: Maybe FilePath
, optStatistics :: Bool
}

defaultOptions :: Options
defaultOptions = Options {
  optEvalFile = Nothing
, optNPredictions = 1000
, optPredictor = "rforest"
, optSaveState = Nothing
, optLoadState = Nothing
, optHistogram = Nothing
, optStatistics = False
}

options :: [OptDescr (Options -> Options)]
options =
  [
    Option "e" []
      (ReqArg (\ e opts -> opts { optEvalFile = Just e }) "<eval>")
      "file containing theorems for which we want predictions"
  , Option "n" []
      (ReqArg (\ n opts -> opts { optNPredictions = read n }) "<i>")
      "number of predictions to output"
  , Option "p" []
      (ReqArg (\ p opts -> opts { optPredictor = p }) "<method>")
      "predictor and predictor flags"
  , Option "s" []
      (ReqArg (\ f opts -> opts { optSaveState = Just f }) "<file>")
      "save predictor state"
  , Option "l" []
      (ReqArg (\ f opts -> opts { optLoadState = Just f }) "<file>")
      "load predictor state"
  , Option "h" []
      (ReqArg (\ f opts -> opts { optHistogram = Just f }) "<file>")
      "save feature histogram"
  , Option "t" []
      (NoArg (\ opts -> opts {optStatistics = True})) "print statistics"
  ]

header :: String -> String
header progName = intercalate "\n"
  [ "Usage: " ++ progName ++ " <ftrs> <deps> <seq> [OPTION]..."
  , ""
  , "Mandatory arguments:"
  , "  <ftrs>   file containing the features of theorems"
  , "  <deps>   file containing the dependencies of theorems"
  , "  <seq>    file containing the order of theorems"
  , ""
  , "Optional arguments:"
  ]

showUsage :: IO ()
showUsage = do
  progName <- getProgName
  hPutStr stderr (usageInfo (header progName) options)

main :: IO ()
main = do
  args <- getArgs
  (ftrsF, depsF, seqF, opts) <- case getOpt Permute options args of
    (o, [ff, df, sf], []) ->
      return (ff, df, sf, foldl (flip id) defaultOptions o)
    _ -> showUsage >> exitFailure

  lblMap         <- I.loadSequenceFile seqF
  lbls           <- I.loadLabelsFile lblMap depsF
  (ftrs, ftrMap) <- I.loadFeaturesFile lblMap ftrsF

  let fT = SampleFeatures ftrs
  let learnData = LearnData {
    sampleLabels   = lbls
  , sampleFeatures = fT
  , labelBM        = lblMap
  , featureBM      = ftrMap
  , nPredictions   = optNPredictions opts
  }

  F.mapM_ (`writeFeatureHistogram` fT) $ optHistogram opts

  when (optStatistics opts) $ putStr $ showStats learnData

  let (pID, pOpts) = parsePredictorOptions $ optPredictor opts
      operation = maybe interaction evaluation (optEvalFile opts)
      run (predictor, predictorCfg) =
        runPredictorT (operation predictor opts) learnData predictorCfg

  case pID of
    "rforest" -> run $ initialiseForest     learnData pOpts
    "knn"     -> run $ initialiseKNN        learnData pOpts
    "nbayes"  -> run $ initialiseNaiveBayes learnData pOpts
    _         -> error "unknown predictor"
