{-# LANGUAGE DeriveGeneric #-}

module Statistics (main) where

import           Prelude hiding (mapM)
import           Control.Arrow ((&&&), second)
import           Control.Monad (liftM, when)
import           Data.Foldable (toList)
import qualified Data.ByteString.Char8 as BC
import qualified Data.IntMap as IM
import qualified Data.IntSet as IS
import           Data.List (findIndices, inits, sortBy, transpose)
import           Data.Maybe (mapMaybe)
import           Data.Ord (comparing)
import qualified Data.Sequence as Seq
import           Data.Traversable (mapM)
import           GHC.Generics (Generic)
import           Options.Applicative
import           Text.Printf (printf)

import qualified Interpreter as I
import           Table
import           Types
import           TypeUtils
import           Utils


data SampleStatistics = SampleStatistics {
  auc       :: Double
, cover     :: Double
, rank      :: Double
, precision :: Int
, recall    :: Int
, coverI    :: [Double]
, recallI   :: [Int]
, lblRanks  :: [(LabelID, Maybe Int)]
, sampleID  :: SampleID
} deriving Generic

data FileStatistics = FileStatistics {
  aucF       :: Double
, coverF     :: Double
, rankF      :: Double
, precisionF :: Double
, recallF    :: Double
, coverIF    :: [Double]
, recallIF   :: [Int]
, lblRanksF  :: [(SampleID, [(LabelID, Maybe Int)])]
, nSamples   :: Int
}


-- -----------------------------------------------------------------------------
-- Sample statistics

calculateAUC :: Int -> Int -> [Int] -> Double
calculateAUC nLabels nPreds hitIndices
  | nLabels == 0 || nPreds == 0 = 1
  | otherwise = fromIntegral (integrateDiracs nPreds hitIndices) /
                fromIntegral (nLabels * nPreds)

calculateRank :: Int -> Int -> [Int] -> Double
calculateRank nDeps nPreds hitIndices =
  fromIntegral (sum hitIndices + missed * nPreds) / fromIntegral nDeps
  where missed = nDeps - length hitIndices

calculateRecall :: Int -> Int -> [Int] -> Int
calculateRecall nDeps nPreds hitIndices
  | length hitIndices == nDeps = 1 + last hitIndices
  | otherwise                  = 1 + nPreds

calculateCoverI :: Int -> [Int] -> [Int] -> [Double]
calculateCoverI nLabels hitIndices =
  map (\ i -> missed i / fromIntegral (min nLabels i))
  where missed i = fromIntegral $ length (takeWhile (< i) hitIndices)


-- | Compare actual labels with predicted labels.
evaluateSample :: SampleID -> Labels -> Labels -> Maybe SampleStatistics
evaluateSample _ []   _     = Nothing
evaluateSample _ _    []    = Nothing
evaluateSample s lbls preds = Just SampleStatistics {
    auc       = calculateAUC nLabels nPreds hitIndices
  , cover     = fromIntegral (length lblsInPreds) / nLabelsN
  , rank      = calculateRank nLabels nPreds hitIndices
  , precision = length predsInLbls
  , recall    = recall'
  , coverI    = calculateCoverI nLabels hitIndices indices
  , recallI   = map (\ i -> if i >= recall' then 1 else 0) indices
  , lblRanks  = map (\ l -> (l, IM.lookup l preds100M)) lbls
  , sampleID  = s
  }
  where
    nLabels   = length lbls
    nPreds    = length preds
    nLabelsN  = fromIntegral nLabels
    preds100  = take 100 preds
    preds100M = IM.fromList $ zip preds100 [1 ..]
    lblsS     = IS.fromList lbls

    lblsInPreds = filter (`IM.member` preds100M) lbls
    predsInLbls = filter (`IS.member` lblsS) preds100
    hitIndices  = findIndices (`IS.member` lblsS) preds

    indices = [1 .. 200]
    --indices = [2^x | x <- [0 .. 11]]

    recall'  = calculateRecall nLabels nPreds hitIndices


-- -----------------------------------------------------------------------------
-- File statistics

evaluateSamples :: [(SampleID, (Labels, Labels))] -> [SampleStatistics]
evaluateSamples = mapMaybe (\ (s, (l, p)) -> evaluateSample s l p)

{-
filterFuture :: SampleID -> Labels -> Labels
filterFuture (SampleID s) = filter (< s)
-}

-- | Consider only the first n predicted labels.
cutoff :: Int -> SampleLabels -> SampleLabels
cutoff n (SampleLabels preds) = SampleLabels $ fmap (take n) preds

-- | Match labels with predictions.
match :: SampleLabels -> SampleLabels -> [(SampleID, (Labels, Labels))]
match (SampleLabels lbls) (SampleLabels preds) =
  toList $ Seq.mapWithIndex (\ s lp -> (SampleID s, lp)) $ Seq.zip lbls preds

combineStats :: [SampleStatistics] -> FileStatistics
combineStats stats = FileStatistics {
    aucF       = sum (map auc       stats) / n
  , coverF     = sum (map cover     stats) / n
  , rankF      = sum (map rank      stats) / n
  , precisionF = fromIntegral (sum (map precision stats)) / n
  , recallF    = fromIntegral (sum (map recall    stats)) / n
  , coverIF    = map (100 / n *) $ sumLists $ map coverI stats
  , recallIF   = sumLists $ map recallI stats
  , lblRanksF  = map (sampleID &&& lblRanks) stats
  , nSamples   = nI
  }
  where nI    = length stats
        n     = fromIntegral nI

-- | Calculate statistics for increasing number of samples.
iterativeStats :: [SampleStatistics] -> [FileStatistics]
iterativeStats = map combineStats . tail . inits


-- -----------------------------------------------------------------------------
-- Printing

statisticsList :: FilePath -> FileStatistics -> [String]
statisticsList fp stats =
  fp : show (nSamples stats) : zipWith (\ f d -> printf ("%." ++ show (d :: Int) ++ "f") $ f stats)
  [coverF, precisionF, recallF, aucF, rankF]
  [     3,          3,       2,    4,     2]

showLblRanks :: LabelBM -> FileStatistics -> [String]
showLblRanks lblMap =
  map (\ (s, r) -> unlines $ (showSampleS s ++ ":") : showRanks r) . lblRanksF
  where
    showSampleS = BC.unpack . showSample lblMap
    showLabelS = BC.unpack . showLabel lblMap
    showRanks = map (uncurry showRank) . sortBy (comparing snd)
    showRank lbl rnk = "  " ++
      maybe "N/A" (printf "%3d") rnk ++ " -> " ++ showLabelS lbl


-- -----------------------------------------------------------------------------
-- Command-line options

data PredictionFormat = Predict LabelBM | FastXML Samples

data Options = Options {
  optCSV     :: Bool
, optCoverI  :: Bool
, optRecallI :: Bool
, optRanks   :: Bool
, optIterate :: Bool
, optCutoff  :: Maybe Int
, optFastXML :: Maybe FilePath
}

data Arguments = Arguments {
  _sequenceFile :: FilePath
, _labelsFile :: FilePath
, _optins :: Options
, _predictionFiles :: [FilePath]
}

optParser :: Parser Options
optParser = Options
  <$> switch (long "csv"       <> help "output tables in CSV format")
  <*> switch (long "cover-i"   <> help "print Cover-i")
  <*> switch (long "recall-i"  <> help "print Recall-i")
  <*> switch (long "ranks"     <> help "print prediction ranks")
  <*> switch (long "iterative" <> help "iterative statistics")
  <*> optional (option auto
      ( long "cutoff" <> metavar "N"
      <> help "consider only the first N predicted labels"))
  <*> optional (strOption
      ( long "fastxml" <> metavar "EVAL"
      <> help "interpret predictions in FastXML format, with additional evaluation file"))

argParser :: Parser Arguments
argParser = Arguments
  <$> argument str ( metavar "SEQ"
                   <> help "file containing the order of theorems")
  <*> argument str ( metavar "DEPS"
                   <> help "file containing the dependencies of theorems")
  <*> optParser
  <*> many (argument str (metavar "predictions"))

parserInfo :: ParserInfo Arguments
parserInfo = info (helper <*> argParser)
  ( fullDesc
  <> progDesc "Compare prediction files with actual dependencies.")


-- -----------------------------------------------------------------------------
-- Prediction loading

type Loaders = ([BC.ByteString] -> SampleLabels, FilePath -> IO SampleLabels)

loadFormat :: PredictionFormat -> Loaders
loadFormat (Predict m) = (I.loadLabels &&& I.loadLabelsFile) m
loadFormat (FastXML e) = (I.loadFastXMLPredictions &&& I.loadFastXMLPredictionsFile) e

-- | Load predictions either from files or from stdin if file list is empty.
loadPredictions :: Loaders -> [FilePath] -> IO [(FilePath, SampleLabels)]
loadPredictions (loadLine, loadFile) predFPs =
  if null predFPs
  then liftM (\ c -> [("<stdin>", loadLine $ BC.lines c)]) BC.getContents
  else mapM (\ fp -> liftPair (return fp, loadFile fp)) predFPs


-- -----------------------------------------------------------------------------
-- Main

main :: IO ()
main = do
  Arguments seqFP labelsFP opts predFPs <- execParser parserInfo


  lblMap <- I.loadSequenceFile seqFP
  lbls   <- I.loadLabelsFile lblMap labelsFP
  format <- maybe (return $ Predict lblMap)
           (\fp -> I.loadEvaluationFile lblMap fp >>= return . FastXML) $
           optFastXML opts
  preds  <- loadPredictions (loadFormat format) predFPs

  let sampleStats = map (second $ evaluateSamples . match lbls .
                         maybe id cutoff (optCutoff opts)) preds
  let fileStats = (if optIterate opts then iterStats else allStats) sampleStats

{-
  let xx = map (\ (_, SampleLabels sl) -> fmap (showLabels lblMap . take 3) sl) preds
  mapM_ (mapM BC.putStrLn) xx
-}

  let descriptors = ["", "Samples", "Cov", "Prec", "Rec", "AUC", "Rank"]
      statisticsOpts = TableOptions {
        separator = if optCSV opts then ";" else " "
      , cellWidths = maximum (map (length . fst) preds) : replicate 6 8
      , alignments = AlignLeft : repeat AlignRight
      }
      graphOpts = statisticsOpts {
        cellWidths = map (length . fst) preds
      , alignments = repeat AlignRight
      }
      dashes = headerSep statisticsOpts '-'

      stats = descriptors : dashes : map (uncurry statisticsList) fileStats
      covIs = map (\ (fp, s) -> fp : show2fs  (coverIF  s)) fileStats
      recIs = map (\ (fp, s) -> fp : map show (recallIF s)) fileStats
      lRnks = map (\ (fp, s) -> fp : showLblRanks lblMap s) fileStats

  let showStats = not $ optCoverI opts || optRecallI opts || optRanks opts

  when showStats $ putStr $ renderTable statisticsOpts stats
  when (optCoverI  opts) $ putStr $ renderTable graphOpts $ transpose covIs
  when (optRecallI opts) $ putStr $ renderTable graphOpts $ transpose recIs
  when (optRanks   opts) $ mapM_ (putStrLn . unlines) lRnks

  where iterStats = concatMap (\ (fp, p) -> zip (repeat fp) $ iterativeStats p)
        allStats = map (second combineStats)
        show2fs = map (printf "%.2f")
