{-# LANGUAGE OverloadedStrings #-}

module TypeUtils where

import           Control.Arrow (first)
import           Data.ByteString (ByteString)
import qualified Data.ByteString.Char8 as BC
import           Data.Foldable (toList)
import qualified Data.IntMap as IM
import           Data.Tuple (swap)
import qualified Data.Sequence as Seq

import qualified IntBiMap as IBM
import           Types
import           Utils


-- -----------------------------------------------------------------------------
-- Helper functions

nAllLabels :: LearnData -> Int
nAllLabels = IBM.size . labelBM

nAllFeatures :: LearnData -> Int
nAllFeatures = IBM.size . featureBM

sampleLabelsList :: SampleLabels -> [(SampleID, Labels)]
sampleLabelsList (SampleLabels sL) = map (first SampleID) $ zip [0 ..] $ toList sL

labelSamples :: SampleLabels -> LabelMap Samples
labelSamples = intMapFromPairs . map swap . distributeFst . sampleLabelsList

getSampleFeatures :: SampleFeatures -> SampleID -> Features
getSampleFeatures (SampleFeatures fT) (SampleID s) = Seq.index fT s

getSampleLabels :: SampleLabels -> SampleID -> Labels
getSampleLabels (SampleLabels lT) (SampleID s) = Seq.index lT s

-- | Because of the isomorphism between labels and samples, we can convert
-- every sample to a label.
labelOfSample :: SampleID -> LabelID
labelOfSample (SampleID s) = s

sampleOfLabel :: LabelID -> SampleID
sampleOfLabel = SampleID

samplesRange :: SampleID -> SampleID -> Samples
samplesRange (SampleID from) (SampleID to) = map SampleID [from .. to - 1]


-- -----------------------------------------------------------------------------
-- Feature statistics

-- | Update feature frequency information with new samples.
featureFreq :: SampleFeatures -> FeatureFreq -> Samples -> FeatureFreq
featureFreq fT fF =
  foldr (\ f -> IM.insertWith (+) f 1) fF . concatMap (getSampleFeatures fT)

-- | Calculate IDF of a feature from total number of samples and number
-- of appearances of the feature.
featureIDFs :: Int -> FeatureFreq -> FeatureMap Double
featureIDFs nS = IM.map (\ freq -> nSL - log (fromIntegral freq))
  where nSL = log $ fromIntegral nS


-- -----------------------------------------------------------------------------
-- Printing

showSample :: LabelBM -> SampleID -> ByteString
showSample m (SampleID s) = IBM.findElement m s

showLabel :: LabelBM -> LabelID -> ByteString
showLabel = IBM.findElement

showLabels :: LabelBM -> Labels -> ByteString
showLabels lM = BC.unwords . map (showLabel lM)

-- | Show a sample, followed by a list of labels.
showSampleLabels :: LabelBM -> SampleID -> Labels -> ByteString
showSampleLabels lM s ls = BC.concat [showSample lM s, ":", showLabels lM ls]
